using GroupLab.Networking;
using MEdiaSpaceCore;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace MEdiaSpaceCalgary
{
    public partial class CombinedForm : Form
    {
        //
        // Thread-safe UI-updating delegates
        //
        internal delegate void SetPictureBoxImage(PictureBox pbox, Image image);
        internal delegate void SetControlText(Control control, String text);
        internal delegate void SetControlEnbabled(Control control, bool enabled);


        //
        // Locals
        //
        
        // Private status variable
        private bool _talking;

        // Webcam source variable
        private VideoWrapper _videoWrapper;

        // Audio wrapper variable
        private AudioWrapper _audioWrapper;

        // Shared dictionary variables
        private SharedDictionary _sd;
        private Subscription _canmoreUserHandleSub;

        // Networking bits
        private TcpServer _server;


        //
        // Lifecycle
        //
        public CombinedForm()
        {
            InitializeComponent();
        }


        //
        // UI component handlers
        //
        private void CombinedForm_Load(object sender, EventArgs e)
        {
            _talking = false;
            _server = null;

            SetUIEnabled(false);
            SetCallDisabled(true);
            SetCallStatus(false);
            SetLocalImage(null);
            SetRemoteImage(null);

            Variables.maxFPS = Constants.DEFAULT_SLOW_FRAMERATE;
            _videoWrapper = new VideoWrapper();
            _videoWrapper.ImageSent += new VideoWrapper.LocalImageSentHandler(SetLocalImage);
            _videoWrapper.ImageReceived += new VideoWrapper.RemoteImageSentHandler(SetRemoteImage);
            _videoWrapper.ClientClosed += new VideoWrapper.ClientClosedHandler(ClientClosed);
            _videoWrapper.Start();

            string url = "tcp://" + Constants.CALGARY_IP + ":" + Constants.SHARED_DICTIONARY_NAME;
            if (UrlPrompt.Show(ref url) == DialogResult.OK)
            {
                _sd = new SharedDictionary();
                _sd.Url = url;
                _sd.Opened += new EventHandler(sd_Opened);
                _sd.Open(false);
            }
        }

        private void chatInputBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                OnMessageEntered(chatInputBox.Text);
                chatInputBox.Text = "";
            }
        }
        
        private void documentButton_Click(object sender, EventArgs e)
        {
            OnDocumentClicked();
        }

        private void phoneButton_Click(object sender, EventArgs e)
        {
            OnPhoneClicked();
        }

        private void toolsButton_Click(object sender, EventArgs e)
        {
            OnToolsClicked();
        }

        
        //
        // Source/Sink callbacks
        //
        void audioSource_AudioSourceStatusChanged(bool talking)
        {
            _talking = talking;
            SetCallStatus(_talking);

            ClearChatSpool();
        }


        //
        // Shared Dictionary callbacks
        //
        void sd_Opened(object sender, EventArgs e)
        {
            _canmoreUserHandleSub = new Subscription();
            _canmoreUserHandleSub.BeginInit();
            _canmoreUserHandleSub.Pattern = Constants.SDKEY_CANMORE_SKYPE_USERNAME;
            _canmoreUserHandleSub.Dictionary = _sd;
            _canmoreUserHandleSub.Notified += new SubscriptionEventHandler(_canmoreUserHandleSub_Notified);
            _canmoreUserHandleSub.EndInit();

            _audioWrapper = new AudioWrapper(null);
            _sd[Constants.SDKEY_CALGARY_SKYPE_USERNAME] = _audioWrapper.CurrentHandle();
            _audioWrapper.AudioStatusChanged += new AudioWrapper.AudioStatusChangedHandler(audioSource_AudioSourceStatusChanged);
            _audioWrapper.CallAttaching += new AudioWrapper.CallAttachingHandler(SetCallDisabled);

            _sd[Constants.SDKEY_CALGARY_NETWORK_PORT] = Constants.DEFAULT_NETWORK_PORT;
            _server = new TcpServer(Constants.DEFAULT_NETWORK_PORT);
            _server.ClientConnected += new TcpServer.ClientConnectedHandler(_server_ClientConnected);
            Thread serverThread = new Thread(new ThreadStart(_server.Start));
            serverThread.IsBackground = true;
            serverThread.Start();
        }

        void _canmoreUserHandleSub_Notified(object sender, SubscriptionEventArgs e)
        {
            if (_sd.Contains(Constants.SDKEY_CANMORE_SKYPE_USERNAME))
                SetCallDisabled(false);
            else
                SetCallDisabled(true);
        }


        //
        // TcpServer callback
        //
        void _server_ClientConnected(System.Net.Sockets.TcpClient client)
        {
            SetConnected(true);
            SetUIEnabled(true);
            SetCallStatus(false);
            SetRemoteImage(null);
            ClearChatSpool();
           
            _videoWrapper.SetClient(client);
        }


        //
        // Interaction sinks
        //
        private void OnDocumentClicked()
        {
        }

        private void OnPhoneClicked()
        {
            if (!_talking && _sd.Contains(Constants.SDKEY_CALGARY_SKYPE_USERNAME))
            {
                _audioWrapper.SetSinkUsername((String)_sd[Constants.SDKEY_CANMORE_SKYPE_USERNAME]);
                Thread audioSourceThread = new Thread(new ThreadStart(_audioWrapper.StartCall));
                audioSourceThread.IsBackground = true;
                audioSourceThread.Start();
            }
            else
                _audioWrapper.StopCall();
        }

        private void OnToolsClicked()
        {
        }

        private void OnMessageEntered(String message)
        {
            // TEST CODE
            AddChatMessage("Message:", message);
            // END TEST CODE
        }


        //
        // Content distribution functions
        //
        private void SetConnected(bool connected)
        {
            if (connected)
                DisplayTextInControl(statusLabel, Constants.CALGARY_CONNECTED_STRING);
            else
                DisplayTextInControl(statusLabel, Constants.CALGARY_NOT_CONNECTED_STRING);
        }

        private void SetUIEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(documentButton, enabled);
            EnableControl(toolsButton, enabled);
        }

        private void SetCallDisabled(bool disabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(phoneButton, !disabled);
        }

        private void SetCallStatus(bool talking)
        {
            if (talking) {
                phoneButton.Image = global::MEdiaSpaceCalgary.Properties.Resources.redphone_large;
                Variables.maxFPS = Constants.DEFAULT_FAST_FRAMERATE;
            }
            else
            {
                phoneButton.Image = global::MEdiaSpaceCalgary.Properties.Resources.greenphone_large;
                Variables.maxFPS = Constants.DEFAULT_SLOW_FRAMERATE;
            }
        }

        private void SetRemoteImage(Image image)
        {
            DisplayImageInPictureBox(incomingPictureBox, image);
        }

        private void SetLocalImage(Image image)
        {
            if (null != image)
                image.RotateFlip(RotateFlipType.Rotate180FlipY);

            DisplayImageInPictureBox(outgoingPictureBox, image);
        }

        private void ClearChatSpool()
        {
            DisplayTextInControl(chatOutputBox, "");
        }

        private void AddChatMessage(String sender, String message)
        {
            DisplayTextInControl(chatOutputBox, chatOutputBox.Text + sender + ": " + message + System.Environment.NewLine);
        }

        private void ClientClosed()
        {
            SetConnected(false);
            SetUIEnabled(false);
            SetCallStatus(false);
            SetLocalImage(null);
            SetRemoteImage(null);
            ClearChatSpool();

            // Wipe out the old Skype info
            _sd.Remove(Constants.SDKEY_CANMORE_SKYPE_USERNAME);

            // And wait for another client
            Thread serverThread = new Thread(new ThreadStart(_server.Start));
            serverThread.IsBackground = true;
            serverThread.Start();
        }


        //
        // Thread-safe UI updaters
        //
        internal void DisplayImageInPictureBox(PictureBox pbox, Image image)
        {
            if (pbox.InvokeRequired)
            {
                SetPictureBoxImage theDelegate = new SetPictureBoxImage(DisplayImageInPictureBox);
                BeginInvoke(theDelegate, new object[] { pbox, image });
            }
            else
                pbox.Image = image;
        }

        internal void DisplayTextInControl(Control control, String text)
        {
            if (control.InvokeRequired)
            {
                SetControlText theDelegate = new SetControlText(DisplayTextInControl);
                BeginInvoke(theDelegate, new object[] { control, text });
            }
            else
                control.Text = text;
        }

        internal void EnableControl(Control control, bool enabled)
        {
            if (control.InvokeRequired)
            {
                SetControlEnbabled theDelegate = new SetControlEnbabled(EnableControl);
                BeginInvoke(theDelegate, new object[] { control, enabled });
            }
            else
                control.Enabled = enabled;
        }

        private void outgoingPictureBox_Click(object sender, EventArgs e)
        {

        }
    }
}