using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Net.Sockets;
using System.Text;

namespace MEdiaSpaceCore
{
    public class VideoSink
    {
        // Callback bits so the hosting UI can get a copy of images as they're sent out
        public delegate void RemoteImageSentHandler(Image image);
        public event RemoteImageSentHandler ImageReceived;

        // Networking bits (these are provided by the caller!)
        private NetworkStream _stream;

        // Networking parts and pieces
        private byte[] _tinyBuffer;


        public VideoSink(NetworkStream stream)
        {
            _stream = stream;
        }

        public void Start()
        {
            // Establish a one-byte read buffer (just so we know when new data is arriving)
            _tinyBuffer = new byte[1];

            // Set up an asynchronous read loop
            _stream.BeginRead(_tinyBuffer, 0, 1, new AsyncCallback(stream_ReceivedBytes), null);
        }

        public void Stop()
        {
            // TODO: Implement
        }

        public void stream_ReceivedBytes(IAsyncResult asyncResult)
        {
            // Figure out what kind of packet we're looking at...
            switch (_tinyBuffer[0])
            {
                case (byte)Constants.NETWORK_PACKETS.VIDEO:
                    // Capture our timing data
                    // Frame format:
                    // - Image width (4 bytes)
                    // - Image height (4 bytes)
                    // - Image encoding (4 bytes)
                    // - Image data length (4 bytes)
                    // - Image data (variable length)

                    // Read the image width
                    byte[] dataBuffer = new byte[Constants.HEADER_FIELD_LENGTH];
                    _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                    int imageWidth = System.BitConverter.ToInt32(dataBuffer, 0);

                    // Read the image height
                    _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                    int imageHeight = System.BitConverter.ToInt32(dataBuffer, 0);

                    // Read the image encoding
                    _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                    Constants.VIDEO_PROTOCOLS imageEncoding =
                        (Constants.VIDEO_PROTOCOLS)System.BitConverter.ToInt32(dataBuffer, 0);

                    // Read the image data length
                    _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                    int imageSize = System.BitConverter.ToInt32(dataBuffer, 0);

                    // Create an image buffer that's the right size to receive the image object
                    byte[] imageBuffer = new byte[imageSize];
                    int imageBytesRead = 0;

                    // Read the image data
                    while (imageBytesRead < imageSize)
                    {
                        // Read the next chunk of the image into the image buffer
                        imageBytesRead += _stream.Read(imageBuffer, imageBytesRead, imageSize - imageBytesRead);
                    }

                    // Convert the image buffer back into a Bitmap object (different process for each encoding)
                    Bitmap bmap = null;
                    switch (imageEncoding)
                    {
                        case Constants.VIDEO_PROTOCOLS.RAW:
                            unsafe
                            {
                                fixed (byte* imageBufferAddr = imageBuffer)
                                {
                                    IntPtr imageBufferPtr = (IntPtr)imageBufferAddr;
                                    bmap = new Bitmap(imageWidth, imageHeight, imageWidth * Constants.BYTES_PER_PIXEL,
                                                      System.Drawing.Imaging.PixelFormat.Format24bppRgb, imageBufferPtr);
                                }
                            }
                            break;

                        case Constants.VIDEO_PROTOCOLS.JPEG:
                            MemoryStream memStream = new MemoryStream();
                            memStream.Write(imageBuffer, 0, imageSize);
                            bmap = new Bitmap(memStream);
                            memStream.Close();
                            memStream.Dispose();
                            break;

                        default:
                            break;
                    }

                    // If it worked out and we have somewhere to send it, then send it
                    if (bmap != null && ImageReceived != null)
                            ImageReceived(bmap);

                    // In any case, tell the sender that we're ready for the next frame!
                    _stream.WriteByte((byte)Constants.NETWORK_PACKETS.ACK);

                    break;

                case (byte)Constants.NETWORK_PACKETS.AUDIO:
                case (byte)Constants.NETWORK_PACKETS.DATA:
                default:
                    break;
            }

            // And go back to sleep until we receive another packet
            _stream.BeginRead(_tinyBuffer, 0, 1, new AsyncCallback(stream_ReceivedBytes), null);
        }
    }
}
